@@ -1,4 +1,4 @@
-# Copyright (C) 2003-2015 John Goerzen & contributors
+# Copyright (C) 2003-2016 John Goerzen & contributors
 #
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -20,6 +20,7 @@ import os
 import time
 from sys import exc_info
 import traceback
+import six
 
 from offlineimap import mbnames, CustomConfig, OfflineImapError
 from offlineimap import globals
@@ -27,6 +28,8 @@ from offlineimap.repository import Repository
 from offlineimap.ui import getglobalui
 from offlineimap.threadutil import InstanceLimitedThread
 
+FOLDER_NAMESPACE = 'LIMITED_FOLDER_'
+
 try:
     import fcntl
 except:
@@ -35,24 +38,7 @@ except:
 # FIXME: spaghetti code alert!
 def getaccountlist(customconfig):
     # Account names in a list.
-    return customconfig.getsectionlist('Account')
-
-# FIXME: spaghetti code alert!
-def AccountListGenerator(customconfig):
-    """Returns a list of instanciated Account class, one per account name."""
-
-    return [Account(customconfig, accountname)
-            for accountname in getaccountlist(customconfig)]
-
-# FIXME: spaghetti code alert!
-def AccountHashGenerator(customconfig):
-    """Returns a dict of instanciated Account class with the account name as
-    key."""
-
-    retval = {}
-    for item in AccountListGenerator(customconfig):
-        retval[item.getname()] = item
-    return retval
+    return [name.lstrip() for name in customconfig.getsectionlist('Account')]
 
 
 class Account(CustomConfig.ConfigHelperMixin):
@@ -80,10 +66,9 @@ class Account(CustomConfig.ConfigHelperMixin):
         self.name = name
         self.metadatadir = config.getmetadatadir()
         self.localeval = config.getlocaleval()
-        # current :mod:`offlineimap.ui`, can be used for logging:
+        # Current :mod:`offlineimap.ui`, can be used for logging:
         self.ui = getglobalui()
         self.refreshperiod = self.getconffloat('autorefresh', 0.0)
-        # should we run in "dry-run" mode?
         self.dryrun = self.config.getboolean('general', 'dry-run')
         self.quicknum = 0
         if self.refreshperiod == 0.0:
@@ -198,13 +183,31 @@ class Account(CustomConfig.ConfigHelperMixin):
         self.ui.serverdiagnostics(local_repo, 'Local')
         #self.ui.serverdiagnostics(statusrepos, 'Status')
 
+    def deletefolder(self, foldername):
+        remote_repo = Repository(self, 'remote')
+
+        try:
+            if self.dryrun:
+                self.ui.info("would try to remove '%s' on remote of '%s' "
+                    "account"% (foldername, self))
+            else:
+                remote_repo.deletefolder(foldername)
+                self.ui.info("Folder '%s' deleted."% foldername)
+            return 0
+        except Exception as e:
+            self.ui.error(e)
+            return 1
+
 
 class SyncableAccount(Account):
     """A syncable email account connecting 2 repositories.
 
     Derives from :class:`accounts.Account` but contains the additional
     functions :meth:`syncrunner`, :meth:`sync`, :meth:`syncfolders`,
-    used for syncing."""
+    used for syncing.
+
+    In multi-threaded mode, one instance of this object is run per "account"
+    thread."""
 
     def __init__(self, *args, **kwargs):
         Account.__init__(self, *args, **kwargs)
@@ -223,9 +226,12 @@ class SyncableAccount(Account):
             pass
         except IOError:
             self._lockfd.close()
-            raise OfflineImapError("Could not lock account %s. Is another "
-                "instance using this account?"% self,
-                OfflineImapError.ERROR.REPO), None, exc_info()[2]
+            six.reraise(OfflineImapError,
+                        OfflineImapError(
+                            "Could not lock account %s. Is another "
+                            "instance using this account?"% self,
+                            OfflineImapError.ERROR.REPO),
+                        exc_info()[2])
 
     def _unlock(self):
         """Unlock the account, deleting the lock file"""
@@ -239,6 +245,8 @@ class SyncableAccount(Account):
                 pass    # Failed to delete for some reason.
 
     def syncrunner(self):
+        """The target for both single and multi-threaded modes."""
+
         self.ui.registerthread(self)
         try:
             accountmetadata = self.getaccountmeta()
@@ -254,7 +262,7 @@ class SyncableAccount(Account):
                 raise
             return
 
-        # Loop account sync if needed (bail out after 3 failures)
+        # Loop account sync if needed (bail out after 3 failures).
         looping = 3
         while looping:
             self.ui.acct(self)
@@ -321,56 +329,71 @@ class SyncableAccount(Account):
             localrepos = self.localrepos
             statusrepos = self.statusrepos
 
-            # init repos with list of folders, so we have them (and the
-            # folder delimiter etc)
+            # Init repos with list of folders, so we have them (and the
+            # folder delimiter etc).
             remoterepos.getfolders()
             localrepos.getfolders()
 
             remoterepos.sync_folder_structure(localrepos, statusrepos)
-            # replicate the folderstructure between REMOTE to LOCAL
+            # Replicate the folderstructure between REMOTE to LOCAL.
             if not localrepos.getconfboolean('readonly', False):
                 self.ui.syncfolders(remoterepos, localrepos)
 
-            # iterate through all folders on the remote repo and sync
+            # Iterate through all folders on the remote repo and sync.
             for remotefolder in remoterepos.getfolders():
-                # check for CTRL-C or SIGTERM
+                # Check for CTRL-C or SIGTERM.
                 if Account.abort_NOW_signal.is_set(): break
 
                 if not remotefolder.sync_this:
                     self.ui.debug('', "Not syncing filtered folder '%s'"
                                   "[%s]"% (remotefolder, remoterepos))
-                    continue # Ignore filtered folder
+                    continue # Ignore filtered folder.
+
+                # The remote folder names must not have the local sep char in
+                # their names since this would cause troubles while converting
+                # the name back (from local to remote).
+                sep = localrepos.getsep()
+                if (sep != os.path.sep and
+                    sep != remoterepos.getsep() and
+                    sep in remotefolder.getvisiblename()):
+                    self.ui.warn('', "Ignoring folder %s due to unsupported "
+                        "'%s' character serving as local separator."%
+                        (remotefolder.getvisiblename(), localrepos.getsep()))
+                    continue # Ignore unsupported folder name.
+
                 localfolder = self.get_local_folder(remotefolder)
                 if not localfolder.sync_this:
                     self.ui.debug('', "Not syncing filtered folder '%s'"
                                  "[%s]"% (localfolder, localfolder.repository))
-                    continue # Ignore filtered folder
+                    continue # Ignore filtered folder.
+
                 if not globals.options.singlethreading:
-                    thread = InstanceLimitedThread(\
-                        instancename = 'FOLDER_' + self.remoterepos.getname(),
-                        target = syncfolder,
-                        name = "Folder %s [acc: %s]"% (remotefolder.getexplainedname(), self),
-                        args = (self, remotefolder, quick))
+                    thread = InstanceLimitedThread(
+                        limitNamespace="%s%s"% (
+                            FOLDER_NAMESPACE, self.remoterepos.getname()),
+                        target=syncfolder,
+                        name="Folder %s [acc: %s]"% (
+                            remotefolder.getexplainedname(), self),
+                        args=(self, remotefolder, quick)
+                    )
                     thread.start()
                     folderthreads.append(thread)
                 else:
                     syncfolder(self, remotefolder, quick)
-            # wait for all threads to finish
+            # Wait for all threads to finish.
             for thr in folderthreads:
                 thr.join()
-            # Write out mailbox names if required and not in dry-run mode
-            if not self.dryrun:
-                mbnames.write(False)
+            mbnames.writeIntermediateFile(self.name) # Write out mailbox names.
             localrepos.forgetfolders()
             remoterepos.forgetfolders()
         except:
-            #error while syncing. Drop all connections that we have, they
-            #might be bogus by now (e.g. after suspend)
+            # Error while syncing. Drop all connections that we have, they
+            # might be bogus by now (e.g. after suspend).
             localrepos.dropconnections()
             remoterepos.dropconnections()
             raise
         else:
-            # sync went fine. Hold or drop depending on config
+            # Sync went fine. Hold or drop depending on config.
             localrepos.holdordropconnections()
             remoterepos.holdordropconnections()
 
@@ -378,14 +401,14 @@ class SyncableAccount(Account):
         self.callhook(hook)
 
     def callhook(self, cmd):
-        # check for CTRL-C or SIGTERM and run postsynchook
+        # Check for CTRL-C or SIGTERM and run postsynchook.
         if Account.abort_NOW_signal.is_set():
             return
         if not cmd:
             return
         try:
             self.ui.callhook("Calling hook: " + cmd)
-            if self.dryrun: # don't if we are in dry-run mode
+            if self.dryrun:
                 return
             p = Popen(cmd, shell=True,
                       stdin=PIPE, stdout=PIPE, stderr=PIPE,
@@ -398,10 +421,12 @@ class SyncableAccount(Account):
         except Exception as e:
             self.ui.error(e, exc_info()[2], msg="Calling hook")
 
+
 def syncfolder(account, remotefolder, quick):
     """Synchronizes given remote folder for the specified account.
 
-    Filtered folders on the remote side will not invoke this function."""
+    Filtered folders on the remote side will not invoke this function. However,
+    this might be called in a concurrently."""
 
     def check_uid_validity(localfolder, remotefolder, statusfolder):
         # If either the local or the status folder has messages and
@@ -419,7 +444,7 @@ def syncfolder(account, remotefolder, quick):
                 localrepos.restore_atime()
                 return
         else:
-            # Both folders empty, just save new UIDVALIDITY
+            # Both folders empty, just save new UIDVALIDITY.
             localfolder.save_uidvalidity()
             remotefolder.save_uidvalidity()
 
@@ -430,16 +455,15 @@ def syncfolder(account, remotefolder, quick):
         fd.close()
 
     def cachemessagelists_upto_date(localfolder, remotefolder, date):
-        """ Returns messages with uid > min(uids of messages newer than date)."""
+        """Returns messages with uid > min(uids of messages newer than date)."""
 
         localfolder.cachemessagelist(min_date=date)
         check_uid_validity(localfolder, remotefolder, statusfolder)
-        # local messagelist had date restriction applied already. Restrict
+        # Local messagelist had date restriction applied already. Restrict
         # sync to messages with UIDs >= min_uid from this list.
         #
-        # local messagelist might contain new messages (with uid's < 0).
-        positive_uids = filter(
-            lambda uid: uid > 0, localfolder.getmessageuidlist())
+        # Local messagelist might contain new messages (with uid's < 0).
+        positive_uids = [uid for uid in localfolder.getmessageuidlist() if uid > 0]
         if len(positive_uids) > 0:
             remotefolder.cachemessagelist(min_uid=min(positive_uids))
         else:
@@ -451,7 +475,7 @@ def syncfolder(account, remotefolder, quick):
                 min_date=time.gmtime(time.mktime(date) + 24*60*60))
 
     def cachemessagelists_startdate(new, partial, date):
-        """ Retrieve messagelists when startdate has been set for
+        """Retrieve messagelists when startdate has been set for
         the folder 'partial'.
 
         Idea: suppose you want to clone the messages after date in one
@@ -466,8 +490,7 @@ def syncfolder(account, remotefolder, quick):
         might not correspond. But, if we're cloning a folder into a new one,
         [min_uid, ...] does correspond to [1, ...].
 
-        This is just for IMAP-IMAP. For Maildir-IMAP, use maxage instead.
-        """
+        This is just for IMAP-IMAP. For Maildir-IMAP, use maxage instead."""
 
         new.cachemessagelist()
         min_uid = partial.retrieve_min_uid()
@@ -481,9 +504,8 @@ def syncfolder(account, remotefolder, quick):
                 partial.cachemessagelist(min_date=date)
                 # messagelist.keys() instead of getuidmessagelist() because in
                 # the UID mapped case we want the actual local UIDs, not their
-                # remote counterparts
-                positive_uids = filter(
-                    lambda uid: uid > 0, partial.messagelist.keys())
+                # remote counterparts.
+                positive_uids = [uid for uid in list(partial.messagelist.keys()) if uid > 0]
                 if len(positive_uids) > 0:
                     min_uid = min(positive_uids)
                 else:
@@ -503,37 +525,31 @@ def syncfolder(account, remotefolder, quick):
         # Load local folder.
         localfolder = account.get_local_folder(remotefolder)
 
-        # Write the mailboxes
-        mbnames.add(account.name, localfolder.getname(),
-            localrepos.getlocalroot())
+        # Add the folder to the mbnames mailboxes.
+        mbnames.add(account.name, localrepos.getlocalroot(),
+            localfolder.getname())
 
         # Load status folder.
         statusfolder = statusrepos.getfolder(remotefolder.getvisiblename().
             replace(remoterepos.getsep(), statusrepos.getsep()))
-
-        if localfolder.get_uidvalidity() == None:
-            # This is a new folder, so delete the status cache to be
-            # sure we don't have a conflict.
-            # TODO: This does not work. We always return a value, need
-            # to rework this...
-            statusfolder.deletemessagelist()
-
+        statusfolder.openfiles()
         statusfolder.cachemessagelist()
 
-
         # Load local folder.
         ui.syncingfolder(remoterepos, remotefolder, localrepos, localfolder)
 
         # Retrieve messagelists, taking into account age-restriction
-        # options
+        # options.
         maxage = localfolder.getmaxage()
         localstart = localfolder.getstartdate()
         remotestart = remotefolder.getstartdate()
         if (maxage != None) + (localstart != None) + (remotestart != None) > 1:
-            raise OfflineImapError("You can set at most one of the "
-                "following: maxage, startdate (for the local folder), "
-                "startdate (for the remote folder)",
-                OfflineImapError.ERROR.REPO), None, exc_info()[2]
+            six.reraise(OfflineImapError,
+                        OfflineImapError("You can set at most one of the "
+                            "following: maxage, startdate (for the local "
+                            "folder), startdate (for the remote folder)",
+                            OfflineImapError.ERROR.REPO),
+                        exc_info()[2])
         if (maxage != None or localstart or remotestart) and quick:
             # IMAP quickchanged isn't compatible with options that
             # involve restricting the messagelist, since the "quick"
@@ -566,28 +582,28 @@ def syncfolder(account, remotefolder, quick):
             ui.syncingmessages(remoterepos, remotefolder, localrepos, localfolder)
             remotefolder.syncmessagesto(localfolder, statusfolder)
         else:
-            ui.debug('imap', "Not syncing to read-only repository '%s'" \
-                         % localrepos.getname())
+            ui.debug('', "Not syncing to read-only repository '%s'"%
+                    localrepos.getname())
 
-        # Synchronize local changes
+        # Synchronize local changes.
         if not remoterepos.getconfboolean('readonly', False):
             ui.syncingmessages(localrepos, localfolder, remoterepos, remotefolder)
             localfolder.syncmessagesto(remotefolder, statusfolder)
         else:
-            ui.debug('', "Not syncing to read-only repository '%s'" \
-                         % remoterepos.getname())
+            ui.debug('', "Not syncing to read-only repository '%s'"%
+                    remoterepos.getname())
 
         statusfolder.save()
         localrepos.restore_atime()
     except (KeyboardInterrupt, SystemExit):
         raise
     except OfflineImapError as e:
-        # bubble up severe Errors, skip folder otherwise
+        # Bubble up severe Errors, skip folder otherwise.
         if e.severity > OfflineImapError.ERROR.FOLDER:
             raise
         else:
-            ui.error(e, exc_info()[2], msg = "Aborting sync, folder '%s' "
-                     "[acc: '%s']" % (localfolder, account))
+            ui.error(e, exc_info()[2], msg="Aborting sync, folder '%s' "
+                     "[acc: '%s']"% (localfolder, account))
     except Exception as e:
         ui.error(e, msg = "ERROR in syncfolder for %s folder %s: %s"%
             (account, remotefolder.getvisiblename(), traceback.format_exc()))
@@ -595,3 +611,4 @@ def syncfolder(account, remotefolder, quick):
         for folder in ["statusfolder", "localfolder", "remotefolder"]:
             if folder in locals():
                 locals()[folder].dropmessagelistcache()
+        statusfolder.closefiles()
