@@ -1,5 +1,5 @@
 # IMAP server support
-# Copyright (C) 2002 - 2011 John Goerzen & contributors
+# Copyright (C) 2002-2016 John Goerzen & contributors.
 #
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -15,18 +15,23 @@
 #    along with this program; if not, write to the Free Software
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
-from threading import Lock, BoundedSemaphore, Thread, Event, currentThread
 import hmac
 import socket
 import base64
+import json
+import urllib
 import time
 import errno
-from sys import exc_info
+import socket
 from socket import gaierror
+from sys import exc_info
 from ssl import SSLError, cert_time_to_seconds
+from threading import Lock, BoundedSemaphore, Thread, Event, currentThread
+
+import six
 
-from offlineimap import imaplibutil, imaputil, threadutil, OfflineImapError
 import offlineimap.accounts
+from offlineimap import imaplibutil, imaputil, threadutil, OfflineImapError
 from offlineimap.ui import getglobalui
 
 
@@ -39,7 +44,8 @@ try:
 except ImportError:
     pass
 
-class IMAPServer:
+
+class IMAPServer(object):
     """Initializes all variables from an IMAPRepository() instance
 
     Various functions, such as acquireconnection() return an IMAP4
@@ -51,7 +57,10 @@ class IMAPServer:
 
     GSS_STATE_STEP = 0
     GSS_STATE_WRAP = 1
+
     def __init__(self, repos):
+        """:repos: a IMAPRepository instance."""
+
         self.ui = getglobalui()
         self.repos = repos
         self.config = repos.getconfig()
@@ -76,6 +85,13 @@ class IMAPServer:
         self.goodpassword = None
 
         self.usessl = repos.getssl()
+        self.useipv6 = repos.getipv6()
+        if self.useipv6 == True:
+            self.af = socket.AF_INET6
+        elif self.useipv6 == False:
+            self.af = socket.AF_INET
+        else:
+            self.af = socket.AF_UNSPEC
         self.hostname = \
             None if self.preauth_tunnel else repos.gethost()
         self.port = repos.getport()
@@ -87,7 +103,19 @@ class IMAPServer:
         if self.sslcacertfile is None:
             self.__verifycert = None # disable cert verification
         self.fingerprint = repos.get_ssl_fingerprint()
+        self.tlslevel = repos.gettlslevel()
         self.sslversion = repos.getsslversion()
+        self.starttls = repos.getstarttls()
+
+        if self.tlslevel is not "tls_compat" and self.sslversion is None:
+            raise Exception("When 'tls_version' is not 'tls_compat' "
+                "the 'ssl_version' must be set explicitly.")
+
+        self.oauth2_refresh_token = repos.getoauth2_refresh_token()
+        self.oauth2_access_token = repos.getoauth2_access_token()
+        self.oauth2_client_id = repos.getoauth2_client_id()
+        self.oauth2_client_secret = repos.getoauth2_client_secret()
+        self.oauth2_request_url = repos.getoauth2_request_url()
 
         self.delim = None
         self.root = None
@@ -130,6 +158,7 @@ class IMAPServer:
 
     def __getpassword(self):
         """Returns the server password or None"""
+
         if self.goodpassword != None: # use cached good one first
             return self.goodpassword
 
@@ -143,13 +172,6 @@ class IMAPServer:
         self.passworderror = None
         return self.password
 
-    # XXX: is this function used anywhere?
-    def getroot(self):
-        """Returns this server's folder root. Can only be called after one
-        or more calls to acquireconnection."""
-
-        return self.root
-
 
     def releaseconnection(self, connection, drop_conn=False):
         """Releases a connection, returning it to the pool.
@@ -157,7 +179,9 @@ class IMAPServer:
         :param drop_conn: If True, the connection will be released and
            not be reused. This can be used to indicate broken connections."""
 
-        if connection is None: return #noop on bad connection
+        if connection is None:
+            return # Noop on bad connection.
+
         self.connectionlock.acquire()
         self.assignedconnections.remove(connection)
         # Don't reuse broken connections
@@ -190,16 +214,60 @@ class IMAPServer:
 
         authc = self.username
         passwd = self.__getpassword()
-        authz = ''
+        authz = b''
         if self.user_identity != None:
             authz = self.user_identity
-        NULL = u'\x00'
-        retval = NULL.join((authz, authc, passwd)).encode('utf-8')
-        self.ui.debug('imap', '__plainhandler: returning %s' % retval)
+        # At this point all authz, authc and passwd are expected bytes encoded
+        # in UTF-8.
+        NULL = b'\x00'
+        retval = NULL.join((authz, authc, passwd))
+        logsafe_retval = NULL.join((authz, authc, "(passwd hidden for log)"))
+        self.ui.debug('imap', '__plainhandler: returning %s'% logsafe_retval)
         return retval
 
 
-    # XXX: describe function
+    def __xoauth2handler(self, response):
+        if self.oauth2_refresh_token is None \
+                and self.oauth2_access_token is None:
+            return None
+
+        if self.oauth2_access_token is None:
+            if self.oauth2_request_url is None:
+                raise OfflineImapError("No remote oauth2_request_url for "
+                    "repository '%s' specified."%
+                    self, OfflineImapError.ERROR.REPO)
+
+            # Generate new access token.
+            params = {}
+            params['client_id'] = self.oauth2_client_id
+            params['client_secret'] = self.oauth2_client_secret
+            params['refresh_token'] = self.oauth2_refresh_token
+            params['grant_type'] = 'refresh_token'
+
+            self.ui.debug('imap', 'xoauth2handler: url "%s"'%
+                self.oauth2_request_url)
+            self.ui.debug('imap', 'xoauth2handler: params "%s"'% params)
+
+            original_socket = socket.socket
+            socket.socket = self.proxied_socket
+            try:
+                response = urllib.urlopen(
+                    self.oauth2_request_url, urllib.urlencode(params)).read()
+            finally:
+                socket.socket = original_socket
+
+            resp = json.loads(response)
+            self.ui.debug('imap', 'xoauth2handler: response "%s"'% resp)
+            self.oauth2_access_token = resp['access_token']
+
+        self.ui.debug('imap', 'xoauth2handler: access_token "%s"'%
+            self.oauth2_access_token)
+        auth_string = 'user=%s\1auth=Bearer %s\1\1'% (
+            self.username, self.oauth2_access_token)
+        #auth_string = base64.b64encode(auth_string)
+        self.ui.debug('imap', 'xoauth2handler: returning "%s"'% auth_string)
+        return auth_string
+
     def __gssauth(self, response):
         data = base64.b64encode(response)
         try:
@@ -283,6 +351,10 @@ class IMAPServer:
         imapobj.authenticate('PLAIN', self.__plainhandler)
         return True
 
+    def __authn_xoauth2(self, imapobj):
+        imapobj.authenticate('XOAUTH2', self.__xoauth2handler)
+        return True
+
     def __authn_login(self, imapobj):
         # Use LOGIN command, unless LOGINDISABLED is advertized
         # (per RFC 2595)
@@ -314,6 +386,7 @@ class IMAPServer:
         auth_methods = {
           "GSSAPI": (self.__authn_gssapi, False, True),
           "CRAM-MD5": (self.__authn_cram_md5, True, True),
+          "XOAUTH2": (self.__authn_xoauth2, True, True),
           "PLAIN": (self.__authn_plain, True, True),
           "LOGIN": (self.__authn_login, True, False),
         }
@@ -338,7 +411,7 @@ class IMAPServer:
 
             # TLS must be initiated before checking capabilities:
             # they could have been changed after STARTTLS.
-            if tryTLS and not tried_tls:
+            if tryTLS and self.starttls and not tried_tls:
                 tried_tls = True
                 self.__start_tls(imapobj)
 
@@ -367,8 +440,7 @@ class IMAPServer:
 
         if not tried_to_authn:
             methods = ", ".join(map(
-              lambda x: x[5:], filter(lambda x: x[0:5] == "AUTH=",
-               imapobj.capabilities)
+              lambda x: x[5:], [x for x in imapobj.capabilities if x[0:5] == "AUTH="]
             ))
             raise OfflineImapError(u"Repository %s: no supported "
               "authentication mechanisms found; configured %s, "
@@ -412,9 +484,9 @@ class IMAPServer:
 
         # Must be careful here that if we fail we should bail out gracefully
         # and release locks / threads so that the next attempt can try...
-        success = 0
+        success = False
         try:
-            while not success:
+            while success is not True:
                 # Generate a new connection.
                 if self.tunnel:
                     self.ui.connecting('tunnel', self.tunnel)
@@ -423,20 +495,22 @@ class IMAPServer:
                         timeout=socket.getdefaulttimeout(),
                         use_socket=self.proxied_socket,
                         )
-                    success = 1
+                    success = True
                 elif self.usessl:
                     self.ui.connecting(self.hostname, self.port)
                     imapobj = imaplibutil.WrappedIMAP4_SSL(
-                        self.hostname,
-                        self.port,
-                        self.sslclientkey,
-                        self.sslclientcert,
-                        self.sslcacertfile,
-                        self.__verifycert,
-                        self.sslversion,
+                        host=self.hostname,
+                        port=self.port,
+                        keyfile=self.sslclientkey,
+                        certfile=self.sslclientcert,
+                        ca_certs=self.sslcacertfile,
+                        cert_verify_cb=self.__verifycert,
+                        ssl_version=self.sslversion,
                         timeout=socket.getdefaulttimeout(),
                         fingerprint=self.fingerprint,
                         use_socket=self.proxied_socket,
+                        tls_level=self.tlslevel,
+                        af=self.af,
                         )
                 else:
                     self.ui.connecting(self.hostname, self.port)
@@ -444,13 +518,14 @@ class IMAPServer:
                         self.hostname, self.port,
                         timeout=socket.getdefaulttimeout(),
                         use_socket=self.proxied_socket,
+                        af=self.af,
                         )
 
                 if not self.preauth_tunnel:
                     try:
                         self.__authn_helper(imapobj)
                         self.goodpassword = self.password
-                        success = 1
+                        success = True
                     except OfflineImapError as e:
                         self.passworderror = str(e)
                         raise
@@ -502,7 +577,9 @@ class IMAPServer:
                          "'%s'. Make sure you have configured the ser"\
                          "ver name correctly and that you are online."%\
                          (self.hostname, self.repos)
-                raise OfflineImapError(reason, severity), None, exc_info()[2]
+                six.reraise(OfflineImapError,
+                            OfflineImapError(reason, severity),
+                            exc_info()[2])
 
             elif isinstance(e, SSLError) and e.errno == errno.EPERM:
                 # SSL unknown protocol error
@@ -515,7 +592,9 @@ class IMAPServer:
                     reason = "Unknown SSL protocol connecting to host '%s' for "\
                          "repository '%s'. OpenSSL responded:\n%s"\
                          % (self.hostname, self.repos, e)
-                raise OfflineImapError(reason, severity), None, exc_info()[2]
+                six.reraise(OfflineImapError,
+                            OfflineImapError(reason, severity),
+                            exc_info()[2])
 
             elif isinstance(e, socket.error) and e.args[0] == errno.ECONNREFUSED:
                 # "Connection refused", can be a non-existing port, or an unauthorized
@@ -524,14 +603,19 @@ class IMAPServer:
                     "refused. Make sure you have the right host and port "\
                     "configured and that you are actually able to access the "\
                     "network."% (self.hostname, self.port, self.repos)
-                raise OfflineImapError(reason, severity), None, exc_info()[2]
+                six.reraise(OfflineImapError,
+                            OfflineImapError(reason, severity),
+                            exc_info()[2])
             # Could not acquire connection to the remote;
             # socket.error(last_error) raised
             if str(e)[:24] == "can't open socket; error":
-                raise OfflineImapError("Could not connect to remote server '%s' "\
-                    "for repository '%s'. Remote does not answer."
-                    % (self.hostname, self.repos),
-                    OfflineImapError.ERROR.REPO), None, exc_info()[2]
+                six.reraise(OfflineImapError,
+                            OfflineImapError(
+                                "Could not connect to remote server '%s' "
+                                "for repository '%s'. Remote does not answer."%
+                                (self.hostname, self.repos),
+                                OfflineImapError.ERROR.REPO),
+                            exc_info()[2])
             else:
                 # re-raise all other errors
                 raise
@@ -546,7 +630,7 @@ class IMAPServer:
         It's OK if we have maxconnections + 1 or 2 threads, which is what this
         will help us do."""
 
-        self.semaphore.acquire()
+        self.semaphore.acquire() # Blocking until maxconnections has free slots.
         self.semaphore.release()
 
     def close(self):
