@@ -1,5 +1,5 @@
 # IMAP repository support
-# Copyright (C) 2002-2015 John Goerzen & contributors
+# Copyright (C) 2002-2016 John Goerzen & contributors
 #
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -15,15 +15,16 @@
 #    along with this program; if not, write to the Free Software
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
-from threading import Event
 import os
-from sys import exc_info
 import netrc
 import errno
+import codecs
+from sys import exc_info
+from threading import Event
+import six
 
-from offlineimap.repository.Base import BaseRepository
 from offlineimap import folder, imaputil, imapserver, OfflineImapError
-from offlineimap.folder.UIDMaps import MappedIMAPFolder
+from offlineimap.repository.Base import BaseRepository
 from offlineimap.threadutil import ExitNotifyThread
 from offlineimap.utils.distro import get_os_sslcertfile, get_os_sslcertfile_searchpath
 
@@ -31,12 +32,15 @@ from offlineimap.utils.distro import get_os_sslcertfile, get_os_sslcertfile_sear
 class IMAPRepository(BaseRepository):
     def __init__(self, reposname, account):
         """Initialize an IMAPRepository object."""
+
         BaseRepository.__init__(self, reposname, account)
         # self.ui is being set by the BaseRepository
         self._host = None
         self._oauth2_request_url = None
         self.imapserver = imapserver.IMAPServer(self)
         self.folders = None
+        self.copy_ignore_eval = None
+
         # Only set the newmail_hook in an IMAP repository.
         if self.config.has_option(self.getsection(), 'newmail_hook'):
             self.newmail_hook = self.localeval.eval(
@@ -72,10 +76,23 @@ class IMAPRepository(BaseRepository):
     def dropconnections(self):
         self.imapserver.close()
 
+    def get_copy_ignore_UIDs(self, foldername):
+        """Return a list of UIDs to not copy for this foldername."""
+
+        if self.copy_ignore_eval is None:
+            if self.config.has_option(self.getsection(),
+                                     'copy_ignore_eval'):
+                self.copy_ignore_eval = self.localeval.eval(
+                        self.getconf('copy_ignore_eval'))
+            else:
+                self.copy_ignore_eval = lambda x: None
+
+        return self.copy_ignore_eval(foldername)
+
     def getholdconnectionopen(self):
         if self.getidlefolders():
             return 1
-        return self.getconfboolean("holdconnectionopen", 0)
+        return self.getconfboolean("holdconnectionopen", False)
 
     def getkeepalive(self):
         num = self.getconfint("keepalive", 0)
@@ -107,9 +124,12 @@ class IMAPRepository(BaseRepository):
             try:
                 host = self.localeval.eval(host)
             except Exception as e:
-                raise OfflineImapError("remotehosteval option for repository "
-                    "'%s' failed:\n%s"% (self, e), OfflineImapError.ERROR.REPO), \
-                    None, exc_info()[2]
+                six.reraise(OfflineImapError,
+                            OfflineImapError(
+                                "remotehosteval option for repository "
+                                "'%s' failed:\n%s"% (self, e),
+                                OfflineImapError.ERROR.REPO),
+                            exc_info()[2])
             if host:
                 self._host = host
                 return self._host
@@ -128,7 +148,10 @@ class IMAPRepository(BaseRepository):
         (currently -- PLAIN) to inform server about the ID
         we want to authorize as instead of our login name."""
 
-        return self.getconf('remote_identity', default=None)
+        identity = self.getconf('remote_identity', default=None)
+        if identity != None:
+            identity = identity.encode('UTF-8')
+        return identity
 
     def get_auth_mechanisms(self):
         supported = ["GSSAPI", "XOAUTH2", "CRAM-MD5", "PLAIN", "LOGIN"]
@@ -157,13 +180,15 @@ class IMAPRepository(BaseRepository):
 
         if self.config.has_option(self.getsection(), 'remoteusereval'):
             user = self.getconf('remoteusereval')
-        if user != None:
-            return localeval.eval(user)
+            if user != None:
+                return localeval.eval(user).encode('UTF-8')
 
         if self.config.has_option(self.getsection(), 'remoteuser'):
+            # Assume the configuration file to be UTF-8 encoded so we must not
+            # encode this string again.
             user = self.getconf('remoteuser')
-        if user != None:
-            return user
+            if user != None:
+                return user
 
         try:
             netrcentry = netrc.netrc().authenticators(self.gethost())
@@ -195,10 +220,10 @@ class IMAPRepository(BaseRepository):
         return self.getconfint('remoteport', None)
 
     def getipv6(self):
-        return self.getconfboolean('ipv6', None)
+        return self.getconfboolean('ipv6', False)
 
     def getssl(self):
-        return self.getconfboolean('ssl', 1)
+        return self.getconfboolean('ssl', True)
 
     def getsslclientcert(self):
         xforms = [os.path.expanduser, os.path.expandvars, os.path.abspath]
@@ -210,7 +235,7 @@ class IMAPRepository(BaseRepository):
 
     def getsslcacertfile(self):
         """Determines CA bundle.
-        
+
         Returns path to the CA bundle.  It is either explicitely specified
         or requested via "OS-DEFAULT" value (and we will search known
         locations for the current OS and distribution).
@@ -228,6 +253,7 @@ class IMAPRepository(BaseRepository):
 
         xforms = [os.path.expanduser, os.path.expandvars, os.path.abspath]
         cacertfile = self.getconf_xform('sslcacertfile', xforms, None)
+        # Can't use above cacertfile because of abspath.
         if self.getconf('sslcacertfile', None) == "OS-DEFAULT":
             cacertfile = get_os_sslcertfile()
             if cacertfile == None:
@@ -255,6 +281,9 @@ class IMAPRepository(BaseRepository):
     def getsslversion(self):
         return self.getconf('ssl_version', None)
 
+    def getstarttls(self):
+        return self.getconfboolean('starttls', True)
+
     def get_ssl_fingerprint(self):
         """Return array of possible certificate fingerprints.
 
@@ -268,25 +297,36 @@ class IMAPRepository(BaseRepository):
         if self._oauth2_request_url:  # Use cached value if possible.
             return self._oauth2_request_url
 
-        oauth2_request_url = self.getconf('oauth2_request_url', None)
-        if oauth2_request_url != None:
-            self._oauth2_request_url = oauth2_request_url
-            return self._oauth2_request_url
-
-        #raise OfflineImapError("No remote oauth2_request_url for repository "
-            #"'%s' specified."% self, OfflineImapError.ERROR.REPO)
+        self.oauth2_request_url = self.getconf('oauth2_request_url', None)
+        return self._oauth2_request_url
 
     def getoauth2_refresh_token(self):
-        return self.getconf('oauth2_refresh_token', None)
+        refresh_token = self.getconf('oauth2_refresh_token', None)
+        if refresh_token is None:
+            refresh_token = self.localeval.eval(
+                self.getconf('oauth2_refresh_token_eval', "lambda: None"))
+        return refresh_token
 
     def getoauth2_access_token(self):
-        return self.getconf('oauth2_access_token', None)
+        access_token = self.getconf('oauth2_access_token', None)
+        if access_token is None:
+            access_token = self.localeval.eval(
+                self.getconf('oauth2_access_token_eval', "lambda: None"))
+        return access_token
 
     def getoauth2_client_id(self):
-        return self.getconf('oauth2_client_id', None)
+        client_id = self.getconf('oauth2_client_id', None)
+        if client_id is None:
+            client_id = self.localeval.eval(
+                self.getconf('oauth2_client_id_eval', "lambda: None"))
+        return client_id
 
     def getoauth2_client_secret(self):
-        return self.getconf('oauth2_client_secret', None)
+        client_secret = self.getconf('oauth2_client_secret', None)
+        if client_secret is None:
+            client_secret = self.localeval.eval(
+                self.getconf('oauth2_client_secret_eval', "lambda: None"))
+        return client_secret
 
     def getpreauthtunnel(self):
         return self.getconf('preauthtunnel', None)
@@ -298,11 +338,10 @@ class IMAPRepository(BaseRepository):
         return self.getconf('reference', '')
 
     def getdecodefoldernames(self):
-        return self.getconfboolean('decodefoldernames', 0)
+        return self.getconfboolean('decodefoldernames', False)
 
     def getidlefolders(self):
-        localeval = self.localeval
-        return localeval.eval(self.getconf('idlefolders', '[]'))
+        return self.localeval.eval(self.getconf('idlefolders', '[]'))
 
     def getmaxconnections(self):
         num1 = len(self.getidlefolders())
@@ -310,7 +349,7 @@ class IMAPRepository(BaseRepository):
         return max(num1, num2)
 
     def getexpunge(self):
-        return self.getconfboolean('expunge', 1)
+        return self.getconfboolean('expunge', True)
 
     def getpassword(self):
         """Return the IMAP password for this repository.
@@ -329,18 +368,20 @@ class IMAPRepository(BaseRepository):
         # 1. evaluate Repository 'remotepasseval'
         passwd = self.getconf('remotepasseval', None)
         if passwd != None:
-            return self.localeval.eval(passwd)
+            return self.localeval.eval(passwd).encode('UTF-8')
         # 2. read password from Repository 'remotepass'
         password = self.getconf('remotepass', None)
         if password != None:
+            # Assume the configuration file to be UTF-8 encoded so we must not
+            # encode this string again.
             return password
         # 3. read password from file specified in Repository 'remotepassfile'
         passfile = self.getconf('remotepassfile', None)
         if passfile != None:
-            fd = open(os.path.expanduser(passfile))
+            fd = codecs.open(os.path.expanduser(passfile), 'r', 'UTF-8')
             password = fd.readline().strip()
             fd.close()
-            return password
+            return password.encode('UTF-8')
         # 4. read password from ~/.netrc
         try:
             netrcentry = netrc.netrc().authenticators(self.gethost())
@@ -398,7 +439,7 @@ class IMAPRepository(BaseRepository):
             self.imapserver.releaseconnection(imapobj)
         for s in listresult:
             if s == None or \
-                   (isinstance(s, basestring) and s == ''):
+                   (isinstance(s, str) and s == ''):
                 # Bug in imaplib: empty strings in results from
                 # literals. TODO: still relevant?
                 continue
@@ -448,6 +489,19 @@ class IMAPRepository(BaseRepository):
         self.folders = retval
         return self.folders
 
+    def deletefolder(self, foldername):
+        """Delete a folder on the IMAP server."""
+
+        imapobj = self.imapserver.acquireconnection()
+        try:
+            result = imapobj.delete(foldername)
+            if result[0] != 'OK':
+                raise OfflineImapError("Folder '%s'[%s] could not be deleted. "
+                    "Server responded: %s"% (foldername, self, str(result)),
+                    OfflineImapError.ERROR.FOLDER)
+        finally:
+            self.imapserver.releaseconnection(imapobj)
+
     def makefolder(self, foldername):
         """Create a folder on the IMAP server
 
@@ -457,9 +511,12 @@ class IMAPRepository(BaseRepository):
 
         :param foldername: Full path of the folder to be created."""
 
+        if foldername is '':
+            return
+
         if self.getreference():
             foldername = self.getreference() + self.getsep() + foldername
-        if not foldername: # Create top level folder as folder separator
+        if not foldername: # Create top level folder as folder separator.
             foldername = self.getsep()
         self.ui.makefolder(self, foldername)
         if self.account.dryrun:
@@ -476,4 +533,4 @@ class IMAPRepository(BaseRepository):
 
 class MappedIMAPRepository(IMAPRepository):
     def getfoldertype(self):
-        return MappedIMAPFolder
+        return folder.UIDMaps.MappedIMAPFolder
